import opendp.transformations as transformations
import opendp.measurements as measurements
import opendp.combinators as combinators

import json
import builtins
from opendp_logger.constants import PT_TYPE_PREFIX, OPENDP_VERSION

__all__ = ["make_load_json", "make_load_ast"]


def decode_ast(obj):
    if isinstance(obj, dict):
        if obj.get("_type") == "tuple":
            return tuple(decode_ast(i) for i in obj["_items"])

        if obj.get("_type") == "constructor":
            obj = {
                **obj,
                "args": decode_ast(obj["args"]),
                "kwargs": decode_ast(obj["kwargs"]),
            }
            if obj["module"] == "transformations":
                module = transformations
            elif obj["module"] == "measurements":
                module = measurements
            elif obj["module"] == "combinators":
                module = combinators
            else:
                raise ValueError(f"Unrecognized module {obj['module']}.")

            return getattr(module, obj["func"])(*obj["args"], **obj["kwargs"])

        else:
            return {k: decode_ast(v) for k, v in obj.items()}

    if isinstance(obj, list):
        return list(decode_ast(i) for i in obj)

    if isinstance(obj, str) and obj.startswith(PT_TYPE_PREFIX):
        return getattr(builtins, obj[len(PT_TYPE_PREFIX) :])

    return obj


def make_load_json(parse_str: str):
    return make_load_ast(json.loads(parse_str))


def make_load_ast(obj):
    if obj["version"] != OPENDP_VERSION:
        raise ValueError(
            f"OpenDP version in parsed object ({obj['version']}) does not match the current installation ({OPENDP_VERSION})."
        )

    return decode_ast(obj["ast"])



import pkg_resources

PT_TYPE_PREFIX = "py_type:"

try:
    OPENDP_VERSION = pkg_resources.get_distribution("opendp").version
except pkg_resources.DistributionNotFound:
    OPENDP_VERSION = "0.0.0+development"



from opendp_logger import enable_logging, make_load_json
enable_logging()

from opendp.transformations import *
from opendp.measurements import *
from opendp.combinators import *
from opendp.typing import *
from opendp.mod import enable_features

enable_features("contrib")

def test_sized_mean():
    size = 1000
    bounds = (1.0, 10.0)
    meas = make_sized_bounded_mean(size, bounds) >> make_base_laplace(.01)
    serial = meas.to_json()
    make_load_json(serial)

def test_base_laplace():
    meas = make_base_laplace(.01)
    serial = meas.to_json()
    make_load_json(serial)

def test_base_gaussian():
    from opendp.measurements import make_base_gaussian
    from opendp.combinators import make_fix_delta, make_zCDP_to_approxDP
    delta = .000001
    meas = make_fix_delta(
        make_zCDP_to_approxDP(
            make_base_gaussian(scale=10.5, D="VectorDomain<AllDomain<f64>>")), delta)
    print("base gaussian:", meas([80., 90., 100.]))
    assert meas.check(1., (0.6, delta))
    make_load_json(meas.to_json())

def test_composition():
    composed = make_basic_composition([
        make_count(TIA=int, TO=int) >> make_basic_composition([
            make_base_discrete_laplace(scale=2.), 
            make_base_discrete_laplace(scale=200.)
        ]), 
        make_cast_default(int, bool) >> make_cast_default(bool, int) >> make_count(TIA=int, TO=int) >> make_base_discrete_laplace(scale=2.), 
        make_cast_default(int, float) >> make_clamp((0., 10.)) >> make_bounded_sum((0., 10.)) >> make_base_laplace(scale=2.), 

        make_basic_composition([
            make_count(TIA=int, TO=int) >> make_base_discrete_laplace(scale=2.), 
            make_count(TIA=int, TO=float) >> make_base_laplace(scale=2.),
            (
                make_cast_default(int, str) >> 
                make_count_by_categories(categories=["0", "12", "22"]) >> 
                make_base_discrete_laplace(scale=2., D=VectorDomain[AllDomain[int]])
            )
        ])
    ])

    make_load_json(composed.to_json())



import json
import opendp as opendp

from opendp_logger.constants import PT_TYPE_PREFIX, OPENDP_VERSION

import importlib

__all__ = ["enable_logging"]


def wrap_constructor(f_str, f, module_name):
    def wrapped(*args, **kwargs):
        chain = f(*args, **kwargs)
        chain.log = {
            "_type": "constructor",
            "func": f_str,
            "module": module_name,
            "args": args,
            "kwargs": kwargs,
        }
        return chain

    wrapped.__annotations__ = f.__annotations__
    wrapped.__doc__ = f.__doc__

    return wrapped


def to_ast(item):
    if isinstance(item, (opendp.Transformation, opendp.Measurement)):
        if not hasattr(item, "log"):
            msg = "invoke `opendp_logger.enable_logging()` before constructing your measurement"
            raise ValueError(msg)

        return to_ast(item.log)
    if isinstance(item, tuple):
        return {"_type": "tuple", "_items": [to_ast(e) for e in item]}
    if isinstance(item, list):
        return [to_ast(e) for e in item]
    if isinstance(item, dict):
        return {key: to_ast(value) for key, value in item.items()}
    if isinstance(item, type):
        return PT_TYPE_PREFIX + item.__name__
    if isinstance(item, opendp.typing.RuntimeType):
        return str(item)
    return item


def to_json(chain):
    return json.dumps({"version": OPENDP_VERSION, "ast": chain.to_ast()})


enabled = False


def enable_logging():

    global enabled
    if enabled:
        return

    for name in ["transformations", "measurements", "combinators"]:
        module = importlib.import_module(f"opendp.{name}")
        for f in dir(module):
            if f.startswith("make_"):
                module.__dict__[f] = wrap_constructor(f, getattr(module, f), name)

    opendp.Transformation.to_ast = to_ast
    opendp.Measurement.to_ast = to_ast

    opendp.Transformation.to_json = to_json
    opendp.Measurement.to_json = to_json

    enabled = True



from setuptools import find_packages, setup

setup(
    name="opendp-logger",
    version="0.2.0",
    description="A logger wrapper for OpenDP to keep track of, import, export the AST",
    long_description = open("README.md", "r").read(),
    long_description_content_type = "text/markdown",
    url="https://github.com/opendp/opendp-logger",
    author='Oblivious',
    author_email='hello@oblivious.ai',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ], 
    keywords='opendp logger ast',
    packages=find_packages(),
    python_requires=">=3.8, <4",
    install_requires=[
        "opendp >= 0.6.0"
    ],
    package_data={"oblv": ["py.typed"]},
)